# Created: DEC 2013
# Author: Lukas Treyer
# License: GPL

import bpy, os
from bpy.props import StringProperty, BoolProperty, EnumProperty, IntProperty
from .dxfimport.do import Do

bl_info = {
    "name": "Import AutoCAD DXF Format (.dxf)",
    "author": "Lukas Treyer, Manfred Moitzi (support + dxfgrabber library)",
    "version": (0, 8, 4),
    "blender": (2, 7, 0),
    "location": "File > Import > AutoCAD DXF",
    "description": "Import files in the Autocad DXF format (.dxf)",
    "wiki_url": "https://bitbucket.org/treyerl/io_import_scene_dxf/overview",
    "tracker_url": "https://bitbucket.org/treyerl/io_import_scene_dxf/issues?status=new&status=open",
    "support": "OFFICIAL",
    "category": "Import-Export",
    }

epsg_list=[('None', 'None', 'No Coordinate System is available / will be set'),
           ('user', 'User Defined', 'Define the EPSG code'),
           ('EPSG:4326', 'WGS84','World Geodetic System 84; default for lat / lon; EPSG:4326'),
           ('EPSG:3857', 'Spherical Mercator', 'Webbrowser mapping service standard (Google, OpenStreetMap, ESRI); EPSG:3857'),
           ('EPSG:27700', 'National Grid U.K', 'Ordnance Survey National Grid reference system used in Great Britain; EPSG:27700'),
           ('EPSG:2154', 'France (Lambert 93)', 'Lambert Projection for France; EPSG:2154'),
           ('EPSG:5514', 'Czech Republic & Slovakia', 'Coordinate System for Czech Republic and Slovakia; EPSG:5514'),
           ('EPSG:5243', 'LLC Germany', 'Projection for Germany; EPSG:5243'),
           ('EPSG:28992', 'Amersfoort Netherlands', 'Amersfoort / RD New -- Netherlands; EPSG:28992'),
           ('EPSG:21781', 'Swiss CH1903 / LV03', 'Switzerland and Lichtenstein; EPSG:21781'),
           ('EPSG:5880', 'Brazil Polyconic', 'Cartesian 2D; Central, South America; EPSG:5880 '),
           ('EPSG:42103', 'LCC USA', 'Lambert Conformal Conic Projection; EPSG:42103'),
           ('EPSG:3350', 'Russia: Pulkovo 1942 / CS63 zone C0', 'Russian Federation - onshore and offshore; EPSG:3350'),
           ('EPSG:22293', 'Cape / Lo33 South Africa', 'South Africa; EPSG:22293'),
           ('EPSG:27200', 'NZGD49 / New Zealand Map Grid', 'NZGD49 / New Zealand Map Grid; EPSG:27200'),
           ('EPSG:3112', 'GDA94 Australia Lambert', 'GDA94 / Geoscience Australia Lambert; EPSG:3112'),
           ('EPSG:24378', 'India zone I', 'Kalianpur 1975 / India zone I; EPSG:24378'),
           ('EPSG:2326', 'Hong Kong 1980 Grid System', 'Hong Kong 1980 Grid System; EPSG:2326'),
           ('EPSG:3414', 'SVY21 / Singapore TM', 'SVY21 / Singapore TM; EPSG:3414'),
           ]

__version__ = '.'.join([str(s) for s in bl_info['version']])

BY_LAYER = 0
BY_DXFTYPE = 1
SEPARATED = 2
LINKED_OBJECTS = 3
GROUP_INSTANCES = 4

T_Merge = True
T_ImportText = True
T_ImportLight = True
T_ExportAcis = False
T_MergeLines = True
T_OutlinerGroups = True
T_Bbox = True
T_CreateNewScene = False
T_Recenter = False
T_ThicknessBevel = False
T_but_group_by_att = True

def read(filename, obj_merge=BY_LAYER, import_text=True, import_light=True, export_acis=True, merge_lines=True,
         do_bbox=True, block_rep=LINKED_OBJECTS, new_scene=None, recenter=False, projDXF=None, projSCN=None,
         thicknessWidth=True, but_group_by_att=True):
    # import dxf and export nurbs types to sat/sab files 
    # because that's how autocad stores nurbs types in a dxf...
    do = Do(filename, obj_merge, import_text, import_light, export_acis, merge_lines, do_bbox, block_rep, recenter,
            projDXF, projSCN, thicknessWidth, but_group_by_att)
    errors = do.entities(os.path.basename(filename).replace(".dxf", ""), new_scene)

    # display errors
    if len(errors) > 0:
        for error in errors:
            bpy.ops.error.message('INVOKE_DEFAULT', message=error)

    # inform the user about the sat/sab files
    if len(do.acis_files) > 0:
        bpy.ops.error.message('INVOKE_DEFAULT',
                              message="Exported %d NURBS objects to sat/sab files next to your DXF file."
                                      % len(do.acis_files))

def display_groups_in_outliner():
    outliners = (a for a in bpy.context.screen.areas if a.type == "OUTLINER")
    for outliner in outliners:
        outliner.spaces[0].display_mode = "GROUPS"

class MessageOperator(bpy.types.Operator):
    bl_idname = "error.message"
    bl_label = "Message"
    message = StringProperty()
 
    def execute(self, context):
        self.report({'INFO'}, self.message)
        return {'FINISHED'}
 
    def invoke(self, context, event):
        wm = context.window_manager
        return wm.invoke_props_dialog(self, width=500, height=500)
 
    def draw(self, context):
        print(self.message)
        lines = []
        i = 0
        width = 41
        while len(self.message) > 0:
            lines.append(self.message[:width])
            self.message = self.message[width:]
        for line in lines:
            self.layout.label(line)

class IMPORT_OT_dxf(bpy.types.Operator):
    """Import from DXF file format (.dxf)"""
    bl_idname = "import_scene.dxf"
    bl_description = 'Import from DXF file format (.dxf)'
    bl_label = "Import DXF" +' v.'+ __version__
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_options = {'UNDO'}

    filepath = StringProperty(
            name="input file",
            subtype='FILE_PATH'
            )

    filename_ext = ".dxf"

    filter_glob = StringProperty(
            default="*.dxf",
            options={'HIDDEN'},
            )

    merge = BoolProperty(
            name="Merged Objects",
            description="Merge DXF entities to Blender objects",
            default=T_Merge,
            )

    merge_options = EnumProperty(name="Merge",
            description="Merge multiple DXF entities into one Blender object.",
            items=[('by_type', 'by Layer AND Dxf-Type','Merge DXF entities by type AND layer.'),
                   ('by_layer','by Layer', 'Merge DXF entities of a layer to an object.')],
            default='by_layer',
            )

    merge_lines = BoolProperty(
            name="Combine LINE entities to polygons",
            description="Checks if lines are connect on start or end and merges them to a polygon",
            default=T_MergeLines
            )

    import_text = BoolProperty(
            name="Import Text",
            description="Import DXF Text Entities MTEXT and TEXT",
            default=T_ImportText,
            )

    import_light = BoolProperty(
            name="Import Lights",
            description="Import DXF Text Entity LIGHT",
            default=T_ImportLight
            )

    export_acis = BoolProperty(
            name="Export ACIS Entities",
            description="Export Entities consisting of ACIS code to ACIS .sat/.sab files.",
            default=T_ExportAcis
            )

    outliner_groups = BoolProperty(
            name="Display Groups in Outliner(s)",
            description="Make all outliners in current screen layout show groups",
            default=T_OutlinerGroups
            )

    do_bbox = BoolProperty(
            name="Parent Blocks to Bounding Boxes",
            description="Create a bounding box for blocks with more than one object (faster without)",
            default=T_Bbox
            )

    block_options = EnumProperty(name="Blocks As",
            description="Select the representation of DXF blocks: linked objects or group instances. ",
            items=[('linked_objects', 'linked objects', 'block objects get imported as linked objects'),
                   ('group_instances', 'group instances', 'block objects get imported as group instances')],
            default='linked_objects',
            )

    create_new_scene = BoolProperty(
            name="Import DXF to new scene",
            description="Creates a new scene with the name of the imported file.",
            default=T_CreateNewScene
            )

    recenter = BoolProperty(
            name="Center Geometry",
            description="Moves geometry to the center of the scene",
            default=T_Recenter
            )

    thickness_and_width = BoolProperty(
            name="Line thickness and width",
            description="Map thickness and width of lines to Bevel objects and extrusion attribute",
            default=T_ThicknessBevel
            )

    but_group_by_att = BoolProperty(
            name="But group objects by attributes",
            description="If thickness and width are not imported and 'Merge objects' is on, the objects can be split by attributes",
            default=T_but_group_by_att
            )

    # geo referencing
    epsg_dxf = EnumProperty(name="DXF SRID",
            description="Specifies the coordinate system for the DXF file. Check http://epsg.io",
            items=epsg_list,
            default='None',
            )

    epsg_dxf_user = StringProperty(name="EPSG-Code", default="EPSG")

    epsg_scene = EnumProperty(name="Scn SRID",
            description="Specifies the coordinate system for the Scene. Check http://epsg.io",
            items=epsg_list,
            default='None',
            )

    epsg_scene_user = StringProperty(name="EPSG-Code", default="EPSG")

    def draw(self, context):
        # merge options
        self.layout.label("Merge Options:")
        col = self.layout.box().column()
        col.prop(self, "block_options")
        col.prop(self, "do_bbox")
        col.prop(self, 'merge')
        if self.merge:
            col.prop(self, 'merge_options')
        col.prop(self, "merge_lines")

        # general options
        self.layout.label("Line thickness and width:")
        col = self.layout.box().column()
        col.prop(self, "thickness_and_width")
        if not self.thickness_and_width and self.merge:
            col.prop(self, "but_group_by_att")

        # optional objects
        self.layout.label("Optional Objects:")
        col = self.layout.box().column()
        col.prop(self, 'import_text')
        col.prop(self, 'import_light')
        col.prop(self, 'export_acis')

        # view options
        self.layout.label("View Options:")
        col = self.layout.box().column()
        col.prop(self, "outliner_groups")
        col.prop(self, "create_new_scene")
        col.prop(self, "recenter")
        if not self.create_new_scene and self.recenter:
            if "lat" in bpy.context.scene and "lon" in bpy.context.scene:
                col.label("", icon="ERROR")
                col.label("Current scene has geo reference!")
                col.label("You should not center the import!")

        # geo referencing
        try:
            from pyproj import Proj
            self.layout.label("Geo Referencing")
            col = self.layout.box().column()
            valid_dxf_srid = True

            # EPSG DXF
            col.prop(self, "epsg_dxf")
            if self.epsg_dxf == "user":
                try:
                    Proj(init=self.epsg_dxf_user)
                except:
                    col.alert = True
                    valid_dxf_srid = False
                col.prop(self, "epsg_dxf_user")
                col.alert = False
            col.separator()

            # EPSG SCENE
            # warning if SCN EPSG is not None and DXF EPSG is None
            if self.epsg_scene != 'None' and (not valid_dxf_srid or self.epsg_dxf == 'None'):
                col.alert = True

            # draw the name of scenes EPSG if it has one already
            predefined = False
            if not self.create_new_scene:
                if "SRID" in bpy.context.scene:
                    predefined = True
                    srid = str(bpy.context.scene['SRID'])
                    self.epsg_scene = "user"
                    self.epsg_scene_user = srid
                    for info in epsg_list:
                        if srid == info[0]:
                            srid = info[1]

                    try:
                        Proj(init=self.epsg_scene_user)
                        col.label("Scene SRID: " + str(srid))
                    except Exception as e:
                        print(type(e), e)
                        col.label("Scene SRID not valid; ignored", icon="ERROR")
                else:
                    col.prop(self, "epsg_scene")
            else:
                col.prop(self, "epsg_scene")

            # validate user defined EPSG
            if self.epsg_scene == "user" and not predefined:
                valid = True
                try:
                    Proj(init=self.epsg_scene_user)
                except Exception as e:
                    valid = False
                    col.alert = True
                col.prop(self, "epsg_scene_user")
                if not valid:
                    col.label("", icon="ERROR")
                    col.label("invalid Scene SRID!")

                col.alert = False

            # user info
            if self.epsg_scene != 'None':
                if not valid_dxf_srid:
                    col.label("DXF SRID not valid", icon="ERROR")
                if self.epsg_dxf == 'None':
                    col.label("", icon='ERROR')
                    col.label("DXF SRID must be set, otherwise")
                    if self.epsg_scene == 'user':
                        code = self.epsg_scene_user
                    else:
                        code = self.epsg_scene
                    col.label('scene SRID %s is ignored!' % code )

            # make sure scene EPSG is not None if DXF EPSG is not None
            if self.epsg_scene == 'None' and self.epsg_dxf != 'None':
                self.epsg_scene = self.epsg_dxf

        except Exception as e:
            print("geo referencing: ", e)
            pass

    def execute(self, context):
        merge_map = {"by_layer": BY_LAYER, "by_type": BY_DXFTYPE}
        block_map = {"linked_objects": LINKED_OBJECTS, "group_instances": GROUP_INSTANCES}
        merge_options = SEPARATED
        if self.merge:
            merge_options = merge_map[self.merge_options]
        scene = bpy.context.scene
        if self.create_new_scene:
            scene = bpy.data.scenes.new(os.path.basename(self.filepath).replace(".dxf", ""))

        epsg_dxf = None
        epsg_scn = None
        try:
            from pyproj import Proj
            if self.epsg_dxf != 'None':
                if self.epsg_dxf == 'user':
                    epsg_dxf = Proj(init=self.epsg_dxf_user)
                else:
                    epsg_dxf = Proj(init=self.epsg_dxf)
            if self.epsg_scene != 'None':
                if self.epsg_scene == 'user':
                    epsg_scn = Proj(init=self.epsg_scene_user)
                else:
                    epsg_scn = Proj(init=self.epsg_scene)

        except Exception as e:
            print(e)

        read(self.filepath, merge_options, self.import_text, self.import_light, self.export_acis, self.merge_lines,
             self.do_bbox, block_map[self.block_options], scene, self.recenter, epsg_dxf, epsg_scn,
             self.thickness_and_width, self.but_group_by_att)

        if self.outliner_groups:
            display_groups_in_outliner()

        return {'FINISHED'}

    def invoke(self, context, event):
        wm = context.window_manager
        wm.fileselect_add(self)
        return {'RUNNING_MODAL'}

def menu_func(self, context):
    self.layout.operator(IMPORT_OT_dxf.bl_idname, text="AutoCAD DXF")

def register():
    bpy.utils.register_class(MessageOperator)
    bpy.utils.register_module(__name__)
    bpy.types.INFO_MT_file_import.append(menu_func)
 
def unregister():
    bpy.utils.unregister_module(__name__)
    #bpy.utils.unregister_class(MessageOperator)
    bpy.types.INFO_MT_file_import.remove(menu_func)

if __name__ == "__main__":
    register()